@@ -1,6 +1,7 @@
 #ifndef PG_LOGICAL_COMPAT_H
 #define PG_LOGICAL_COMPAT_H
 
+#include "replication/origin.h"
 #include "utils/varlena.h"
 
 #define WaitLatchOrSocket(latch, wakeEvents, sock, timeout) \
@@ -23,9 +24,6 @@
 #define ExecAlterExtensionStmt(stmt) \
 	ExecAlterExtensionStmt(NULL, stmt)
 
-#define pgl_replorigin_drop(roident) \
-	replorigin_drop(roident, true)
-
 /*
  * Pg 11 adds an argument here.  We don't need to special-case 2ndQPostgres
  * anymore because it adds a separate ExecBRDeleteTriggers2 now, so this only
@@ -54,16 +52,22 @@
 
 #define makeDefElem(name, arg) makeDefElem(name, arg, -1)
 
-#define PGLstandard_ProcessUtility(pstmt, queryString, context, params, queryEnv, dest, sentToRemote, qc) \
+#define PGLstandard_ProcessUtility(pstmt, queryString, readOnlyTree, context, params, queryEnv, dest, sentToRemote, qc) \
 	standard_ProcessUtility(pstmt, queryString, context, params, queryEnv, dest, qc)
 
-#define PGLnext_ProcessUtility_hook(pstmt, queryString, context, params, queryEnv, dest, sentToRemote, qc) \
+#define PGLnext_ProcessUtility_hook(pstmt, queryString, readOnlyTree, context, params, queryEnv, dest, sentToRemote, qc) \
 	next_ProcessUtility_hook(pstmt, queryString, context, params, queryEnv, dest, qc)
 
 #define PGLCreateTrigger(stmt, queryString, relOid, refRelOid, constraintOid, indexOid, isInternal) \
 	CreateTrigger(stmt, queryString, relOid, refRelOid, constraintOid, indexOid, InvalidOid, InvalidOid, NULL, isInternal, false);
 
-#define	PGLDoCopy(stmt, queryString, processed) DoCopy(NULL, stmt, -1, 0, processed)
+#define	PGLDoCopy(stmt, queryString, processed) \
+	do \
+	{ \
+		ParseState* pstate = make_parsestate(NULL); \
+		DoCopy(pstate, stmt, -1, 0, processed); \
+		free_parsestate(pstate); \
+	} while (false);
 
 #define PGLReplicationSlotCreate(name, db_specific, persistency) ReplicationSlotCreate(name, db_specific, persistency)
 
@@ -110,4 +114,15 @@
 #define pg_plan_queries(querytrees, query_string, cursorOptions, boundParams) \
 	pg_plan_queries(querytrees, cursorOptions, boundParams)
 
+/* cd142e032ebd50ec7974b3633269477c2c72f1cc removed replorigin_drop */
+inline static void
+replorigin_drop_by_name(char *name, bool missing_ok, bool nowait)
+{
+	RepOriginId	originid;
+
+	originid = replorigin_by_name(name, missing_ok);
+	if (originid != InvalidRepOriginId)
+		replorigin_drop(originid, nowait);
+}
+
 #endif
