@@ -531,6 +531,8 @@ handle_insert(StringInfo s)
 	PGLogicalRelation  *rel;
 	bool				started_tx = ensure_transaction();
 
+	PushActiveSnapshot(GetTransactionSnapshot());
+
 	errcallback_arg.action_name = "INSERT";
 	xact_action_counter++;
 
@@ -541,6 +543,8 @@ handle_insert(StringInfo s)
 	if (!should_apply_changes_for_rel(rel->nspname, rel->relname))
 	{
 		pglogical_relation_close(rel, NoLock);
+		PopActiveSnapshot();
+		CommandCounterIncrement();
 		return;
 	}
 
@@ -596,6 +600,9 @@ handle_insert(StringInfo s)
 		LockRelationIdForSession(&lockid, RowExclusiveLock);
 		pglogical_relation_close(rel, NoLock);
 
+		PopActiveSnapshot();
+		CommandCounterIncrement();
+
 		apply_api.on_commit();
 
 		handle_queued_message(ht, started_tx);
@@ -615,7 +622,12 @@ handle_insert(StringInfo s)
 //			CommitTransactionCommand();
 	}
 	else
+	{
 		pglogical_relation_close(rel, NoLock);
+
+		PopActiveSnapshot();
+		CommandCounterIncrement();
+	}
 }
 
 static void
@@ -654,6 +666,8 @@ handle_update(StringInfo s)
 
 	multi_insert_finish();
 
+	PushActiveSnapshot(GetTransactionSnapshot());
+
 	rel = pglogical_read_update(s, RowExclusiveLock, &hasoldtup, &oldtup,
 								&newtup);
 	errcallback_arg.rel = rel;
@@ -662,12 +676,17 @@ handle_update(StringInfo s)
 	if (!should_apply_changes_for_rel(rel->nspname, rel->relname))
 	{
 		pglogical_relation_close(rel, NoLock);
+		PopActiveSnapshot();
+		CommandCounterIncrement();
 		return;
 	}
 
 	apply_api.do_update(rel, hasoldtup ? &oldtup : &newtup, &newtup);
 
 	pglogical_relation_close(rel, NoLock);
+
+	PopActiveSnapshot();
+	CommandCounterIncrement();
 }
 
 static void
@@ -683,6 +702,8 @@ handle_delete(StringInfo s)
 
 	multi_insert_finish();
 
+	PushActiveSnapshot(GetTransactionSnapshot());
+
 	rel = pglogical_read_delete(s, RowExclusiveLock, &oldtup);
 	errcallback_arg.rel = rel;
 
@@ -690,12 +711,17 @@ handle_delete(StringInfo s)
 	if (!should_apply_changes_for_rel(rel->nspname, rel->relname))
 	{
 		pglogical_relation_close(rel, NoLock);
+		PopActiveSnapshot();
+		CommandCounterIncrement();
 		return;
 	}
 
 	apply_api.do_delete(rel, &oldtup);
 
 	pglogical_relation_close(rel, NoLock);
+
+	PopActiveSnapshot();
+	CommandCounterIncrement();
 }
 
 inline static bool
@@ -1488,6 +1514,7 @@ execute_sql_command_error_cb(void *arg)
 void
 pglogical_execute_sql_command(char *cmdstr, char *role, bool isTopLevel)
 {
+	const char *save_debug_query_string = debug_query_string;
 	List	   *commands;
 	ListCell   *command_i;
 #ifdef PGXC
@@ -1600,7 +1627,7 @@ pglogical_execute_sql_command(char *cmdstr, char *role, bool isTopLevel)
 	if (error_context_stack == &errcallback)
 		error_context_stack = errcallback.previous;
 
-	debug_query_string = NULL;
+	debug_query_string = save_debug_query_string;
 }
 
 /*
