@@ -120,17 +120,24 @@ pglogical_apply_heap_commit(void)
 
 
 static List *
-UserTableUpdateOpenIndexes(EState *estate, TupleTableSlot *slot)
+UserTableUpdateOpenIndexes(ResultRelInfo *relinfo, EState *estate, TupleTableSlot *slot, bool update)
 {
 	List	   *recheckIndexes = NIL;
 
-	if (estate->es_result_relation_info->ri_NumIndices > 0)
+	if (relinfo->ri_NumIndices > 0)
 	{
-		recheckIndexes = ExecInsertIndexTuples(slot,
+		recheckIndexes = ExecInsertIndexTuples(
+#if PG_VERSION_NUM >= 140000
+											   relinfo,
+#endif
+											   slot,
 #if PG_VERSION_NUM < 120000
 											   &slot->tts_tuple->t_self,
 #endif
 											   estate
+#if PG_VERSION_NUM >= 140000
+											   , update
+#endif
 #if PG_VERSION_NUM >= 90500
 											   , false, NULL, NIL
 #endif
@@ -142,7 +149,7 @@ UserTableUpdateOpenIndexes(EState *estate, TupleTableSlot *slot)
 			StringInfoData si;
 			ListCell *lc;
 			const char *idxname, *relname, *nspname;
-			Relation target_rel = estate->es_result_relation_info->ri_RelationDesc;
+			Relation target_rel = relinfo->ri_RelationDesc;
 
 			relname = RelationGetRelationName(target_rel);
 			nspname = get_namespace_name(RelationGetNamespace(target_rel));
@@ -252,7 +259,15 @@ init_apply_exec_state(PGLogicalRelation *rel)
 
 	/* Initialize the executor state. */
 	aestate->estate = create_estate_for_relation(rel->rel, true);
-	aestate->resultRelInfo = aestate->estate->es_result_relation_info;
+
+	aestate->resultRelInfo = makeNode(ResultRelInfo);
+	InitResultRelInfo(aestate->resultRelInfo, rel->rel, 1, 0);
+
+#if PG_VERSION_NUM < 140000
+	aestate->estate->es_result_relations = aestate->resultRelInfo;
+	aestate->estate->es_num_result_relations = 1;
+	aestate->estate->es_result_relation_info = aestate->resultRelInfo;
+#endif
 
 	aestate->slot = ExecInitExtraTupleSlot(aestate->estate);
 	ExecSetSlotDescriptor(aestate->slot, RelationGetDescr(rel->rel));
@@ -313,9 +328,6 @@ pglogical_apply_heap_insert(PGLogicalRelation *rel, PGLogicalTupleData *newtup)
 	ExecSetSlotDescriptor(localslot, RelationGetDescr(rel->rel));
 #endif
 
-	/* Get snapshot */
-	PushActiveSnapshot(GetTransactionSnapshot());
-
 	ExecOpenIndices(aestate->resultRelInfo
 #if PG_VERSION_NUM >= 90500
 					, false
@@ -327,7 +339,7 @@ pglogical_apply_heap_insert(PGLogicalRelation *rel, PGLogicalTupleData *newtup)
 	 * only normal columns. This doesn't just check the replica identity index,
 	 * but it'll prefer it and use it first.
 	 */
-	conflicts_idx_id = pglogical_tuple_find_conflict(aestate->estate,
+	conflicts_idx_id = pglogical_tuple_find_conflict(aestate->resultRelInfo,
 													 newtup,
 													 localslot);
 
@@ -356,7 +368,6 @@ pglogical_apply_heap_insert(PGLogicalRelation *rel, PGLogicalTupleData *newtup)
 		if (aestate->slot == NULL)		/* "do nothing" */
 #endif
 		{
-			PopActiveSnapshot();
 			finish_apply_exec_state(aestate);
 			return;
 		}
@@ -424,7 +435,6 @@ pglogical_apply_heap_insert(PGLogicalRelation *rel, PGLogicalTupleData *newtup)
 				if (aestate->slot == NULL)		/* "do nothing" */
 #endif
 				{
-					PopActiveSnapshot();
 					finish_apply_exec_state(aestate);
 					return;
 				}
@@ -455,8 +465,10 @@ pglogical_apply_heap_insert(PGLogicalRelation *rel, PGLogicalTupleData *newtup)
 							   TTS_TUP(aestate->slot));
 			if (!HeapTupleIsHeapOnly(TTS_TUP(aestate->slot)))
 #endif
-				recheckIndexes = UserTableUpdateOpenIndexes(aestate->estate,
-															aestate->slot);
+				recheckIndexes = UserTableUpdateOpenIndexes(aestate->resultRelInfo,
+															aestate->estate,
+															aestate->slot,
+															true);
 
 			/* AFTER ROW UPDATE Triggers */
 #if PG_VERSION_NUM >= 120000
@@ -482,7 +494,7 @@ pglogical_apply_heap_insert(PGLogicalRelation *rel, PGLogicalTupleData *newtup)
 #else
 		simple_heap_insert(rel->rel, TTS_TUP(aestate->slot));
 #endif
-		UserTableUpdateOpenIndexes(aestate->estate, aestate->slot);
+		UserTableUpdateOpenIndexes(aestate->resultRelInfo, aestate->estate, aestate->slot, false);
 
 		/* AFTER ROW INSERT Triggers */
 #if PG_VERSION_NUM >= 120000
@@ -494,7 +506,6 @@ pglogical_apply_heap_insert(PGLogicalRelation *rel, PGLogicalTupleData *newtup)
 #endif
 	}
 
-	PopActiveSnapshot();
 	finish_apply_exec_state(aestate);
 
 	CommandCounterIncrement();
@@ -526,10 +537,8 @@ pglogical_apply_heap_update(PGLogicalRelation *rel, PGLogicalTupleData *oldtup,
 	ExecSetSlotDescriptor(localslot, RelationGetDescr(rel->rel));
 #endif
 
-	PushActiveSnapshot(GetTransactionSnapshot());
-
 	/* Search for existing tuple with same key */
-	found = pglogical_tuple_find_replidx(aestate->estate, oldtup, localslot,
+	found = pglogical_tuple_find_replidx(aestate->resultRelInfo, oldtup, localslot,
 										 &replident_idx_id);
 
 	/*
@@ -579,7 +588,6 @@ pglogical_apply_heap_update(PGLogicalRelation *rel, PGLogicalTupleData *oldtup,
 			if (aestate->slot == NULL)		/* "do nothing" */
 #endif
 			{
-				PopActiveSnapshot();
 				finish_apply_exec_state(aestate);
 				return;
 			}
@@ -655,8 +663,10 @@ pglogical_apply_heap_update(PGLogicalRelation *rel, PGLogicalTupleData *oldtup,
 								, false
 #endif
 							   );
-				recheckIndexes = UserTableUpdateOpenIndexes(aestate->estate,
-															aestate->slot);
+				recheckIndexes = UserTableUpdateOpenIndexes(aestate->resultRelInfo,
+															aestate->estate,
+															aestate->slot,
+															true);
 			}
 
 			/* AFTER ROW UPDATE Triggers */
@@ -689,7 +699,6 @@ pglogical_apply_heap_update(PGLogicalRelation *rel, PGLogicalTupleData *oldtup,
 	}
 
 	/* Cleanup. */
-	PopActiveSnapshot();
 	finish_apply_exec_state(aestate);
 
 	CommandCounterIncrement();
@@ -715,9 +724,7 @@ pglogical_apply_heap_delete(PGLogicalRelation *rel, PGLogicalTupleData *oldtup)
 	ExecSetSlotDescriptor(localslot, RelationGetDescr(rel->rel));
 #endif
 
-	PushActiveSnapshot(GetTransactionSnapshot());
-
-	if (pglogical_tuple_find_replidx(aestate->estate, oldtup, localslot,
+	if (pglogical_tuple_find_replidx(aestate->resultRelInfo, oldtup, localslot,
 									 &replident_idx_id))
 	{
 		if (aestate->resultRelInfo->ri_TrigDesc &&
@@ -733,9 +740,7 @@ pglogical_apply_heap_delete(PGLogicalRelation *rel, PGLogicalTupleData *oldtup)
 
 			if (!dodelete)		/* "do nothing" */
 			{
-				PopActiveSnapshot();
 				finish_apply_exec_state(aestate);
-				pglogical_relation_close(rel, NoLock);
 				return;
 			}
 		}
@@ -760,7 +765,6 @@ pglogical_apply_heap_delete(PGLogicalRelation *rel, PGLogicalTupleData *oldtup)
 	}
 
 	/* Cleanup. */
-	PopActiveSnapshot();
 	finish_apply_exec_state(aestate);
 
 	CommandCounterIncrement();
@@ -906,6 +910,9 @@ pglogical_apply_heap_mi_flush(void)
 #endif
 			recheckIndexes =
 				ExecInsertIndexTuples(
+#if PG_VERSION_NUM >= 140000
+									  resultRelInfo,
+#endif
 #if PG_VERSION_NUM >= 120000
 									  pglmistate->buffered_tuples[i],
 #else
@@ -914,6 +921,9 @@ pglogical_apply_heap_mi_flush(void)
 #endif
 									  pglmistate->aestate->estate
 #if PG_VERSION_NUM >= 90500
+#if PG_VERSION_NUM >= 140000
+									  , false
+#endif
 									  , false, NULL, NIL
 #endif
 									 );
@@ -1011,13 +1021,16 @@ pglogical_apply_heap_mi_add_tuple(PGLogicalRelation *rel,
 		ExecConstraints(aestate->resultRelInfo, slot,
 						aestate->estate);
 
-	pglmistate->buffered_tuples[pglmistate->nbuffered_tuples++] =
 #if PG_VERSION_NUM >= 120000
-		slot
+	if (pglmistate->buffered_tuples[pglmistate->nbuffered_tuples] == NULL)
+		pglmistate->buffered_tuples[pglmistate->nbuffered_tuples] = table_slot_create(rel->rel, NULL);
+	else
+		ExecClearTuple(pglmistate->buffered_tuples[pglmistate->nbuffered_tuples]);
+	ExecCopySlot(pglmistate->buffered_tuples[pglmistate->nbuffered_tuples], slot);
 #else
-		remotetuple
+	pglmistate->buffered_tuples[pglmistate->nbuffered_tuples] = remotetuple;
 #endif
-		;
+	pglmistate->nbuffered_tuples++;
 	MemoryContextSwitchTo(oldctx);
 }
 
@@ -1035,6 +1048,12 @@ pglogical_apply_heap_mi_finish(PGLogicalRelation *rel)
 
 	finish_apply_exec_state(pglmistate->aestate);
 
+#if PG_VERSION_NUM >= 120000
+	for (int i = 0; i < pglmistate->maxbuffered_tuples; i++)
+		if (pglmistate->buffered_tuples[i])
+			ExecDropSingleTupleTableSlot(pglmistate->buffered_tuples[i]);
+#endif
+
 	pfree(pglmistate->buffered_tuples);
 	pfree(pglmistate);
 
